@@ -17,8 +17,30 @@
 #ifndef NEO4J_UTIL_H
 #define NEO4J_UTIL_H
 
+#include <arpa/inet.h>
+#include <cypher-parser.h>
+#include <netdb.h>
 #include <stdlib.h>
 
+
+#ifndef MAXSERVNAMELEN
+#  ifdef NI_MAXSERV
+#    define MAXSERVNAMELEN NI_MAXSERV
+#  else
+#    define MAXSERVNAMELEN 32
+#  endif
+#endif
+
+#ifndef MAXHOSTNAMELEN
+#  ifdef NI_MAXHOST
+#    define MAXHOSTNAMELEN NI_MAXHOST
+#  else
+#    define MAXHOSTNAMELEN 1025
+#  endif
+#endif
+
+#define NEO4J_MAXHOSTLEN (MAXHOSTNAMELEN + 1 + MAXSERVNAMELEN)
+
 char *strncpy_alloc(char **dest, size_t *cap, const char *s, size_t n);
 
 #endif/*NEO4J_UTIL_H*/
