@@ -23,6 +23,8 @@
 #include <stddef.h>
 
 
+static int parse_failure_message(struct neo4j_failure_details *details,
+        const char *message, neo4j_mpool_t *mpool);
 static struct neo4j_statement_execution_step *meta_execution_steps(
         neo4j_value_t map, const char *description, const char *key_name,
         neo4j_mpool_t *mpool, neo4j_logger_t *logger);
@@ -60,7 +62,7 @@ const neo4j_value_t *neo4j_validate_metadata(const neo4j_value_t *fields,
     if (field_type != NEO4J_MAP)
     {
         neo4j_log_error(logger, "invalid field in %s: got %s, expected MAP",
-                description, neo4j_type_str(field_type));
+                description, neo4j_typestr(field_type));
         errno = EPROTO;
         return NULL;
     }
@@ -69,8 +71,25 @@ const neo4j_value_t *neo4j_validate_metadata(const neo4j_value_t *fields,
 }
 
 
-int neo4j_meta_failure_details(const char **code, const char **message, const
-        neo4j_value_t map, neo4j_mpool_t *mpool, const char *description,
+void neo4j_metadata_log(neo4j_logger_t *logger, uint_fast8_t level,
+        const char *msg, neo4j_value_t metadata)
+{
+    char detail[1024];
+    size_t n = neo4j_ntostring(metadata, detail, sizeof(detail));
+    if (n >= sizeof(detail))
+    {
+        // TODO: dynamically allocate `detail` if static size is insufficient
+        detail[sizeof(detail)-1] = '\0';
+        detail[sizeof(detail)-2] = '.';
+        detail[sizeof(detail)-3] = '.';
+        detail[sizeof(detail)-4] = '.';
+    }
+    neo4j_log(logger, level, "%s: %s", msg, detail);
+}
+
+
+int neo4j_meta_failure_details(struct neo4j_failure_details *details,
+        const neo4j_value_t map, neo4j_mpool_t *mpool, const char *description,
         neo4j_logger_t *logger)
 {
     size_t pdepth = neo4j_mpool_depth(*mpool);
@@ -88,8 +107,18 @@ int neo4j_meta_failure_details(const char **code, const char **message, const
         goto failure;
     }
 
-    *code = code_string;
-    *message = message_string;
+    details->code = code_string;
+    details->message = message_string;
+    // The message contains a lot of detail that needs to be parsed out.
+    // Hopefully neo4j will start to provide these as separate attributes in
+    // the response, avoiding the need for parsing and the risk created
+    // due to the dependence on the specific format of the string.
+    // Ref: https://github.com/neo4j/neo4j/issues/7318
+    if (parse_failure_message(details, message_string, mpool))
+    {
+        goto failure;
+    }
+
     return 0;
 
     int errsv;
@@ -101,6 +130,83 @@ failure:
 }
 
 
+int parse_failure_message(struct neo4j_failure_details *details,
+        const char *s, neo4j_mpool_t *mpool)
+{
+    details->description = s;
+    details->line = 0;
+    details->column = 0;
+    details->offset = 0;
+    details->context = NULL;
+    details->context_offset = 0;
+
+    // Assume that messages containing position information always include
+    // it at the end of the first line as " (line xx, column yy (offset zz))\n"
+    const char *position = strstr(s, " (line ");
+    if (position == NULL)
+    {
+        return 0;
+    }
+
+    unsigned int line;
+    unsigned int column;
+    unsigned int offset;
+    if (sscanf(position, " (line %u, column %u (offset: %u))",
+                &line, &column, &offset) != 3)
+    {
+        return 0;
+    }
+
+    const char *pend = strchr(position, '\n');
+    if (pend == NULL)
+    {
+        return 0;
+    }
+
+    const char *context = pend + 1;
+    if (*context != '"')
+    {
+        return 0;
+    }
+
+    const char *cend = strchr(context, '\n');
+    if (cend == NULL || (cend - context) <= 2 || *(cend-1) != '"')
+    {
+        return 0;
+    }
+
+    size_t coffset = strspn(cend+1, " ");
+    if (coffset == 0 || *(cend+1+coffset) != '^')
+    {
+        return 0;
+    }
+
+    ++context;
+    --cend;
+    --coffset;
+
+    char *description = alloc_string(neo4j_ustring(s, position - s), mpool);
+    if (description == NULL)
+    {
+        return -1;
+    }
+
+    char *ctx = alloc_string(neo4j_ustring(context, cend - context), mpool);
+    if (ctx == NULL)
+    {
+        return -1;
+    }
+
+    details->description = description;
+    details->line = line;
+    details->column = column;
+    details->offset = offset;
+    details->context = ctx;
+    details->context_offset = coffset;
+    return 0;
+}
+
+
 int neo4j_meta_fieldnames(const char * const **names, unsigned int *nnames,
         neo4j_value_t map, neo4j_mpool_t *mpool, const char *description,
         neo4j_logger_t *logger)
@@ -448,7 +554,7 @@ struct neo4j_statement_execution_step *meta_execution_steps(
                 neo4j_log_error(logger,
                         "invalid field in %s: %s is %s, expected Map",
                         description, subpath,
-                        neo4j_type_str(neo4j_type(child)));
+                        neo4j_typestr(neo4j_type(child)));
                 errno = EPROTO;
                 goto failure;
             }
@@ -496,8 +602,8 @@ int map_get_typed(neo4j_value_t *value, neo4j_value_t map, const char *path,
         neo4j_log_error(logger,
                 "invalid field in %s: '%s%s%s' is %s, expected %s",
                 description, (path != NULL)? path : "",
-                (path != NULL)? "." : "", key, neo4j_type_str(neo4j_type(val)),
-                neo4j_type_str(expected));
+                (path != NULL)? "." : "", key, neo4j_typestr(neo4j_type(val)),
+                neo4j_typestr(expected));
         errno = EPROTO;
         return -1;
     }
@@ -599,7 +705,7 @@ int extract_string_list(const char * const **strings, unsigned int *nstrings,
                     "invalid field in %s: %s%s%s[%d] is %s, expected String",
                     description, (path != NULL)? path : "",
                     (path != NULL)? "." : "", key, i,
-                    neo4j_type_str(neo4j_type(sv)));
+                    neo4j_typestr(neo4j_type(sv)));
             errno = EPROTO;
             goto failure;
         }
