@@ -46,6 +46,14 @@ const char *neo4j_error_message(neo4j_result_stream_t *results)
 }
 
 
+const struct neo4j_failure_details *neo4j_failure_details(
+        neo4j_result_stream_t *results)
+{
+    REQUIRE(results != NULL, NULL);
+    return results->failure_details(results);
+}
+
+
 unsigned int neo4j_nfields(neo4j_result_stream_t *results)
 {
     REQUIRE(results != NULL, -1);
@@ -156,8 +164,7 @@ struct run_result_stream
     struct neo4j_statement_plan *statement_plan;
     struct neo4j_update_counts update_counts;
     int failure;
-    const char *error_code;
-    const char *error_message;
+    struct neo4j_failure_details failure_details;
     unsigned int nfields;
     const char *const *fields;
     result_record_t *records;
@@ -171,6 +178,8 @@ static run_result_stream_t *run_rs_open(neo4j_session_t *session);
 static int run_rs_check_failure(neo4j_result_stream_t *self);
 static const char *run_rs_error_code(neo4j_result_stream_t *self);
 static const char *run_rs_error_message(neo4j_result_stream_t *self);
+const struct neo4j_failure_details *run_rs_failure_details(
+        neo4j_result_stream_t *results);
 static unsigned int run_rs_nfields(neo4j_result_stream_t *results);
 static const char *run_rs_fieldname(neo4j_result_stream_t *self,
         unsigned int index);
@@ -187,7 +196,7 @@ static neo4j_value_t run_result_field(const neo4j_result_t *self,
 static neo4j_result_t *run_result_retain(neo4j_result_t *self);
 static void run_result_release(neo4j_result_t *self);
 
-static void notify_session_ending(neo4j_job_t *job);
+static void abort_job(neo4j_job_t *job, int err);
 static int run_callback(void *cdata, neo4j_message_type_t type,
         const neo4j_value_t *argv, uint16_t argc);
 static int pull_all_callback(void *cdata, neo4j_message_type_t type,
@@ -292,18 +301,21 @@ failure:
 
 run_result_stream_t *run_rs_open(neo4j_session_t *session)
 {
-    run_result_stream_t *results = neo4j_calloc(session->config->allocator,
+    assert(session != NULL);
+    neo4j_config_t *config = neo4j_session_config(session);
+
+    run_result_stream_t *results = neo4j_calloc(config->allocator,
             NULL, 1, sizeof(run_result_stream_t));
 
     results->session = session;
-    results->logger = neo4j_get_logger(session->config, "results");
-    results->allocator = session->config->allocator;
-    results->mpool = neo4j_std_mpool(session->config);
-    results->record_mpool = neo4j_std_mpool(session->config);
+    results->logger = neo4j_get_logger(config, "results");
+    results->allocator = config->allocator;
+    results->mpool = neo4j_std_mpool(config);
+    results->record_mpool = neo4j_std_mpool(config);
     results->statement_type = -1;
     results->refcount = 1;
 
-    results->job.notify_session_ending = notify_session_ending;
+    results->job.abort = abort_job;
     if (neo4j_attach_job(session, &(results->job)))
     {
         neo4j_log_debug_errno(results->logger,
@@ -315,6 +327,7 @@ run_result_stream_t *run_rs_open(neo4j_session_t *session)
     result_stream->check_failure = run_rs_check_failure;
     result_stream->error_code = run_rs_error_code;
     result_stream->error_message = run_rs_error_message;
+    result_stream->failure_details = run_rs_failure_details;
     result_stream->nfields = run_rs_nfields;
     result_stream->fieldname = run_rs_fieldname;
     result_stream->fetch_next = run_rs_fetch_next;
@@ -353,7 +366,7 @@ const char *run_rs_error_code(neo4j_result_stream_t *self)
     run_result_stream_t *results = container_of(self,
             run_result_stream_t, _result_stream);
     REQUIRE(results != NULL, NULL);
-    return results->error_code;
+    return results->failure_details.code;
 }
 
 
@@ -362,7 +375,17 @@ const char *run_rs_error_message(neo4j_result_stream_t *self)
     run_result_stream_t *results = container_of(self,
             run_result_stream_t, _result_stream);
     REQUIRE(results != NULL, NULL);
-    return results->error_message;
+    return results->failure_details.message;
+}
+
+
+const struct neo4j_failure_details *run_rs_failure_details(
+        neo4j_result_stream_t *self)
+{
+    run_result_stream_t *results = container_of(self,
+            run_result_stream_t, _result_stream);
+    REQUIRE(results != NULL, NULL);
+    return &(results->failure_details);
 }
 
 
@@ -557,7 +580,8 @@ int run_rs_close(neo4j_result_stream_t *self)
 neo4j_value_t run_result_field(const neo4j_result_t *self,
         unsigned int index)
 {
-    const result_record_t *record = (const result_record_t *)self;
+    const result_record_t *record = container_of(self,
+            result_record_t, _result);
     REQUIRE(record != NULL, neo4j_null);
     return neo4j_list_get(record->list, index);
 }
@@ -581,7 +605,7 @@ void run_result_release(neo4j_result_t *self)
 }
 
 
-void notify_session_ending(neo4j_job_t *job)
+void abort_job(neo4j_job_t *job, int err)
 {
     run_result_stream_t *results = container_of(job,
             run_result_stream_t, job);
@@ -594,7 +618,7 @@ void notify_session_ending(neo4j_job_t *job)
     results->session = NULL;
     if (results->streaming && results->failure == 0)
     {
-        set_failure(results, NEO4J_SESSION_ENDED);
+        set_failure(results, err);
     }
 }
 
@@ -630,8 +654,7 @@ int run_callback(void *cdata, neo4j_message_type_t type,
     }
 
     char description[128];
-    snprintf(description, sizeof(description),
-            "%s message received in %p (in response to RUN)",
+    snprintf(description, sizeof(description), "%s in %p (response to RUN)",
             neo4j_message_type_str(type), (void *)session);
 
     if (type != NEO4J_SUCCESS_MESSAGE)
@@ -649,6 +672,11 @@ int run_callback(void *cdata, neo4j_message_type_t type,
         return -1;
     }
 
+    if (neo4j_log_is_enabled(session->logger, NEO4J_LOG_TRACE))
+    {
+        neo4j_metadata_log(logger, NEO4J_LOG_TRACE, description, *metadata);
+    }
+
     if (neo4j_meta_fieldnames(&(results->fields), &(results->nfields),
                 *metadata, &(results->mpool), description, logger))
     {
@@ -717,6 +745,8 @@ int stream_end(run_result_stream_t *results, neo4j_message_type_t type,
         return 0;
     }
 
+    neo4j_config_t *config = neo4j_session_config(session);
+
     if (type == NEO4J_IGNORED_MESSAGE)
     {
         if (results->failure == 0)
@@ -748,8 +778,7 @@ int stream_end(run_result_stream_t *results, neo4j_message_type_t type,
     }
 
     char description[128];
-    snprintf(description, sizeof(description),
-            "SUCCESS message received in %p (in response to %s)",
+    snprintf(description, sizeof(description), "SUCCESS in %p (response to %s)",
             (void *)session, src_message_type);
 
     const neo4j_value_t *metadata = neo4j_validate_metadata(argv, argc,
@@ -762,9 +791,7 @@ int stream_end(run_result_stream_t *results, neo4j_message_type_t type,
 
     if (neo4j_log_is_enabled(logger, NEO4J_LOG_TRACE))
     {
-        char buf[4096];
-        neo4j_log_trace(logger, "%s SUCCESS metadata: %s", src_message_type,
-                neo4j_tostring(*metadata, buf, sizeof(buf)));
+        neo4j_metadata_log(logger, NEO4J_LOG_TRACE, description, *metadata);
     }
 
     results->statement_type =
@@ -776,7 +803,7 @@ int stream_end(run_result_stream_t *results, neo4j_message_type_t type,
     }
 
     results->statement_plan = neo4j_meta_plan(*metadata, description,
-            session->config, logger);
+            config, logger);
     if (results->statement_plan == NULL && errno != NEO4J_NO_PLAN_AVAILABLE)
     {
         set_failure(results, errno);
@@ -798,7 +825,6 @@ int await(run_result_stream_t *results, const unsigned int *condition)
 {
     if (*condition > 0 && neo4j_session_sync(results->session, condition))
     {
-        neo4j_log_trace_errno(results->logger, "neo4j_session_sync failed");
         set_failure(results, errno);
         return -1;
     }
@@ -829,7 +855,7 @@ int append_result(run_result_stream_t *results,
         neo4j_log_error(results->logger,
                 "invalid field in RECORD message received in %p"
                 " (got %s, expected List)", (void *)session,
-                neo4j_type_str(arg_type));
+                neo4j_typestr(arg_type));
         errno = EPROTO;
         return -1;
     }
@@ -841,6 +867,9 @@ int append_result(run_result_stream_t *results,
         return 0;
     }
 
+    assert(session != NULL);
+    neo4j_config_t *config = neo4j_session_config(session);
+
     result_record_t *record = neo4j_mpool_calloc(&(results->record_mpool),
             1, sizeof(result_record_t));
     if (record == NULL)
@@ -852,7 +881,7 @@ int append_result(run_result_stream_t *results,
 
     // save memory for the record with the record
     record->mpool = results->record_mpool;
-    results->record_mpool = neo4j_std_mpool(session->config);
+    results->record_mpool = neo4j_std_mpool(config);
 
     record->list = argv[0];
     record->next = NULL;
@@ -910,8 +939,7 @@ int set_eval_failure(run_result_stream_t *results, const char *src_message_type,
     set_failure(results, NEO4J_STATEMENT_EVALUATION_FAILED);
 
     char description[128];
-    snprintf(description, sizeof(description),
-            "FAILURE message received in %p (in response to %s)",
+    snprintf(description, sizeof(description), "FAILURE in %p (response to %s)",
             (void *)(results->session), src_message_type);
 
     const neo4j_value_t *metadata = neo4j_validate_metadata(argv, argc,
@@ -922,9 +950,14 @@ int set_eval_failure(run_result_stream_t *results, const char *src_message_type,
         return -1;
     }
 
-    if (neo4j_meta_failure_details(&(results->error_code),
-                &(results->error_message), *metadata, &(results->mpool),
-                description, results->logger))
+    if (neo4j_log_is_enabled(results->logger, NEO4J_LOG_TRACE))
+    {
+        neo4j_metadata_log(results->logger, NEO4J_LOG_TRACE, description,
+                *metadata);
+    }
+
+    if (neo4j_meta_failure_details(&(results->failure_details), *metadata,
+                &(results->mpool), description, results->logger))
     {
         set_failure(results, errno);
         return -1;
@@ -941,6 +974,5 @@ void set_failure(run_result_stream_t *results, int error)
     results->failure = error;
     results->streaming = false;
     results->awaiting_records = 0;
-    results->error_code = NULL;
-    results->error_message = NULL;
+    memset(&(results->failure_details), 0, sizeof(results->failure_details));
 }
