@@ -16,10 +16,8 @@
  */
 #include "../../config.h"
 #include "session.h"
-#include "connection.h"
 #include "iostream.h"
 #include "memory.h"
-#include "messages.h"
 #include "metadata.h"
 #include "serialization.h"
 #include "util.h"
@@ -32,23 +30,22 @@ static_assert(NEO4J_REQUEST_ARGV_PREALLOC >= 2,
 
 static int session_start(neo4j_session_t *session);
 static int session_clear(neo4j_session_t *session);
+static int session_reset(neo4j_session_t *session);
+
 static int send_requests(neo4j_session_t *session);
 static int receive_responses(neo4j_session_t *session,
-        const unsigned int *condition);
+        const unsigned int *condition, bool interruptable);
 static int drain_queued_requests(neo4j_session_t *session);
 
 static struct neo4j_request *new_request(neo4j_session_t *session);
 static void pop_request(neo4j_session_t* session);
 
-static int initialize(neo4j_session_t *session);
+static int initialize(neo4j_session_t *session, unsigned int attempts);
 static int initialize_callback(void *cdata, neo4j_message_type_t type,
         const neo4j_value_t *argv, uint16_t argc);
 static int ack_failure(neo4j_session_t *session);
 static int ack_failure_callback(void *cdata, neo4j_message_type_t type,
        const neo4j_value_t *argv, uint16_t argc);
-static int reset(neo4j_session_t *session);
-static int reset_callback(void *cdata, neo4j_message_type_t type,
-       const neo4j_value_t *argv, uint16_t argc);
 
 
 neo4j_session_t *neo4j_new_session(neo4j_connection_t *connection)
@@ -69,8 +66,9 @@ neo4j_session_t *neo4j_new_session(neo4j_connection_t *connection)
             (void *)session, (void *)connection);
 
     session->connection = connection;
-    session->config = connection->config;
     session->logger = logger;
+    neo4j_atomic_bool_init(&(session->processing), false);
+    neo4j_atomic_bool_init(&(session->reset_requested), false);
     if (session_start(session))
     {
         goto failure;
@@ -93,6 +91,18 @@ failure:
 }
 
 
+static inline bool interrupted(neo4j_session_t *session)
+{
+    return neo4j_atomic_bool_get(&(session->reset_requested));
+}
+
+
+neo4j_connection_t *neo4j_session_connection(neo4j_session_t *session)
+{
+    return session->connection;
+}
+
+
 int session_start(neo4j_session_t *session)
 {
     if (neo4j_attach_session(session->connection, session))
@@ -108,7 +118,7 @@ int session_start(neo4j_session_t *session)
     assert(session->request_queue_size > 0);
     assert(session->request_queue_depth == 0);
 
-    if (initialize(session))
+    if (initialize(session, 0))
     {
         assert(session->request_queue_depth <= 1);
         session->request_queue_depth = 0;
@@ -126,29 +136,51 @@ failure:
 }
 
 
-static int session_clear(neo4j_session_t *session)
+/**
+ * End all jobs in the session and drain all requests
+ *
+ * @internal
+ *
+ * @param [session] The session.
+ * @return 0 on success, <0 on error (errno will be set).
+ */
+int session_clear(neo4j_session_t *session)
 {
     REQUIRE(session != NULL, -1);
-    REQUIRE(session->connection != NULL, -1);
+    if (session->connection == NULL)
+    {
+        errno = NEO4J_SESSION_ENDED;
+        return -1;
+    }
+    if (neo4j_atomic_bool_set(&(session->processing), true))
+    {
+        errno = NEO4J_SESSION_BUSY;
+        return -1;
+    }
+
     int err = 0;
     int errsv = errno;
 
+    // notify all jobs first, so that they can handle received
+    // responses appropriately
     for (neo4j_job_t *job = session->jobs; job != NULL;)
     {
-        neo4j_job_notify_session_ending(job);
+        neo4j_job_abort(job, NEO4J_SESSION_ENDED);
         neo4j_job_t *next = job->next;
         job->next = NULL;
         job = next;
     }
     session->jobs = NULL;
 
-    if (!session->failed && receive_responses(session, NULL))
+    // Receive responses to inflight requests
+    if (!session->failed && receive_responses(session, NULL, false))
     {
         err = -1;
         errsv = errno;
         session->failed = true;
     }
 
+    // drain any remaining requests
     if (drain_queued_requests(session) && err == 0)
     {
         err = -1;
@@ -157,6 +189,88 @@ static int session_clear(neo4j_session_t *session)
     }
     assert(session->request_queue_depth == 0);
 
+    neo4j_atomic_bool_set(&(session->processing), false);
+    errno = errsv;
+    return err;
+}
+
+
+int session_reset(neo4j_session_t *session)
+{
+    assert(session != NULL);
+    assert(session->connection != NULL);
+
+    const neo4j_config_t *config = neo4j_session_config(session);
+    neo4j_mpool_t mpool =
+            neo4j_mpool(config->allocator, config->mpool_block_size);
+    int err = 0;
+    int errsv = errno;
+
+    // notify all jobs first, so that they can handle received
+    // responses appropriately
+    for (neo4j_job_t *job = session->jobs; job != NULL;)
+    {
+        neo4j_job_abort(job, NEO4J_SESSION_RESET);
+        neo4j_job_t *next = job->next;
+        job->next = NULL;
+        job = next;
+    }
+    session->jobs = NULL;
+
+    // process any already inflight requests
+    if (receive_responses(session, NULL, false) < 0)
+    {
+        err = -1;
+        errsv = errno;
+        session->failed = true;
+        goto cleanup;
+    }
+
+    // receive response to RESET
+    neo4j_message_type_t type;
+    const neo4j_value_t *argv;
+    uint16_t argc;
+    if (neo4j_connection_recv(session->connection, &mpool, &type, &argv, &argc))
+    {
+        neo4j_log_trace_errno(session->logger,
+                "neo4j_connection_recv failed");
+        err = -1;
+        errsv = errno;
+        session->failed = true;
+        goto cleanup;
+    }
+
+    neo4j_log_debug(session->logger, "rcvd %s in response to RESET in %p",
+            neo4j_message_type_str(type), (void *)session);
+
+    if (type != NEO4J_SUCCESS_MESSAGE)
+    {
+        neo4j_log_error(session->logger,
+                "unexpected %s message received in %p"
+                " (expected SUCCESS in response to RESET)",
+                neo4j_message_type_str(type), (void *)session);
+        err = -1;
+        errsv = EPROTO;
+        session->failed = true;
+        goto cleanup;
+    }
+
+cleanup:
+    // ensure queue is empty
+    if (drain_queued_requests(session) && err == 0)
+    {
+        err = -1;
+        errsv = errno;
+        session->failed = true;
+    }
+
+    neo4j_mpool_drain(&mpool);
+
+    if (err == 0)
+    {
+        neo4j_log_debug(session->logger, "session reset (%p)", (void *)session);
+    }
+
     errno = errsv;
     return err;
 }
@@ -186,7 +300,6 @@ int neo4j_end_session(neo4j_session_t *session)
     neo4j_log_debug(session->logger, "session ended (%p)", (void *)session);
 
     session->connection = NULL;
-    session->config = NULL;
     neo4j_logger_release(session->logger);
     session->logger = NULL;
     free(session);
@@ -197,16 +310,47 @@ int neo4j_end_session(neo4j_session_t *session)
 
 int neo4j_reset_session(neo4j_session_t *session)
 {
-    if (session_clear(session))
+    REQUIRE(session != NULL, -1);
+    if (session->connection == NULL)
     {
+        errno = NEO4J_SESSION_ENDED;
         return -1;
     }
-    if (reset(session))
+    if (session->failed)
     {
+        errno = NEO4J_SESSION_FAILED;
         return -1;
     }
-    neo4j_log_debug(session->logger, "session reset (%p)", (void *)session);
-    return 0;
+
+    // immediately send RESET request onto the connection
+    if (neo4j_connection_send(session->connection, NEO4J_RESET_MESSAGE, NULL, 0))
+    {
+        session->failed = true;
+        return -1;
+    }
+
+    neo4j_log_trace(session->logger, "sent RESET in %p", (void *)session);
+
+    // Check and set the reset_requested sentinal and then check if processing
+    // is already taking place.
+    if (neo4j_atomic_bool_set(&(session->reset_requested), true) ||
+        neo4j_atomic_bool_set(&(session->processing), true))
+    {
+        return 0;
+    }
+
+    int err = session_reset(session);
+    // clear reset_requested BEFORE ending processing, to ensure it is not
+    // set if processing resumes
+    neo4j_atomic_bool_set(&(session->reset_requested), false);
+    neo4j_atomic_bool_set(&(session->processing), false);
+    return err;
+}
+
+
+bool neo4j_credentials_expired(const neo4j_session_t *session)
+{
+    return session->credentials_expired;
 }
 
 
@@ -253,6 +397,21 @@ int neo4j_detach_job(neo4j_session_t *session, neo4j_job_t *job)
 }
 
 
+/**
+ * Process requests and responses for a session.
+ *
+ * @internal
+ *
+ * @param [session] The session.
+ * @param [condition] A pointer to a condition flag, that must remain
+ *         greater than zero for processing of responses to continue. This
+ *         allows processing to be stopped when sufficient responses have
+ *         been received to satisfy the current demands. If `NULL`, then
+ *         processing will continue until there are no further outstanding
+ *         requests (or a failure occurs).
+ * @return 0 on success, -1 if a error occurs (errno will be set and all
+ *         requests will be drained).
+ */
 int neo4j_session_sync(neo4j_session_t *session, const unsigned int *condition)
 {
     REQUIRE(session != NULL, -1);
@@ -263,13 +422,25 @@ int neo4j_session_sync(neo4j_session_t *session, const unsigned int *condition)
         errno = NEO4J_SESSION_FAILED;
         return -1;
     }
-
-    while (*condition > 0 && session->request_queue_depth > 0)
+    if (neo4j_atomic_bool_set(&(session->processing), true))
     {
-        int result = receive_responses(session, condition);
+        errno = NEO4J_SESSION_BUSY;
+        return -1;
+    }
+
+    int err = -1;
+
+    while (*condition > 0 && session->request_queue_depth > 0 &&
+            !interrupted(session))
+    {
+        int result = receive_responses(session, condition, true);
         if (result < 0)
         {
-            goto error;
+            goto cleanup;
+        }
+        if (result == 1)
+        {
+            break;
         }
         if (result > 0)
         {
@@ -277,39 +448,66 @@ int neo4j_session_sync(neo4j_session_t *session, const unsigned int *condition)
             if (drain_queued_requests(session))
             {
                 assert(session->request_queue_depth == 0);
+                neo4j_atomic_bool_set(&(session->processing), false);
                 return -1;
             }
             assert(session->request_queue_depth == 0);
+            neo4j_atomic_bool_set(&(session->processing), false);
             return ack_failure(session);
         }
 
         if (send_requests(session))
         {
-            goto error;
+            goto cleanup;
         }
     }
 
-    return 0;
+    if (interrupted(session))
+    {
+        if (!session_reset(session))
+        {
+            errno = NEO4J_SESSION_RESET;
+        }
+        neo4j_atomic_bool_set(&(session->reset_requested), false);
+    }
+    else
+    {
+        err = 0;
+    }
 
     int errsv;
-error:
+cleanup:
     errsv = errno;
-    drain_queued_requests(session);
-    assert(session->request_queue_depth == 0);
+    if (err)
+    {
+        drain_queued_requests(session);
+        assert(session->request_queue_depth == 0);
+    }
+    neo4j_atomic_bool_set(&(session->processing), false);
     errno = errsv;
-    return -1;
+    return err;
 }
 
 
+/**
+ * Send queued requests.
+ *
+ * @internal
+ *
+ * Sends requests, up to the maximum allowed for pipelining by the config.
+ *
+ * @param [session] The session.
+ * @return 0 on success, -1 if an error occurs (errno will be set).
+ */
 int send_requests(neo4j_session_t *session)
 {
     assert(session != NULL);
     neo4j_connection_t *connection = session->connection;
-    const neo4j_config_t *config = session->config;
+    const neo4j_config_t *config = neo4j_session_config(session);
 
     for (unsigned int i = session->inflight_requests;
             i < session->request_queue_depth &&
-            i < config->max_pipelined_requests; ++i)
+            i < config->max_pipelined_requests && !interrupted(session); ++i)
     {
         int offset =
             (session->request_queue_head + i) % session->request_queue_size;
@@ -331,14 +529,34 @@ int send_requests(neo4j_session_t *session)
 }
 
 
-int receive_responses(neo4j_session_t *session, const unsigned int *condition)
+/**
+ * Receive responses to inflight requests.
+ *
+ * @internal
+ *
+ * @param [session] The session.
+ * @param [condition] A pointer to a condition flag, that must remain
+ *         greater than zero for processing of responses to continue. This
+ *         allows processing to be stopped when sufficient responses have
+ *         been received to satisfy the current demands. If `NULL`, then
+ *         processing will continue until there are no further outstanding
+ *         requests (or a failure occurs).
+ * @param [interruptable] If `true`, setting of `reset_requested` will interrupt
+ *         receiving of results.
+ * @return 0 on success, -1 if a error occurs (errno will be set), 1 if
+ *         interrupted, and >1 if a valid FAILURE message is received (in which
+ *         case, all inflight requests will have been drained).
+ */
+int receive_responses(neo4j_session_t *session, const unsigned int *condition,
+        bool interruptable)
 {
     assert(session != NULL);
     ENSURE_NOT_NULL(unsigned int, condition, 1);
     neo4j_connection_t *connection = session->connection;
 
     bool failure = false;
-    while ((failure || *condition > 0) && session->inflight_requests > 0)
+    while ((failure || *condition > 0) && session->inflight_requests > 0 &&
+            (!interruptable || !interrupted(session)))
     {
         neo4j_message_type_t type;
         const neo4j_value_t *argv;
@@ -388,10 +606,28 @@ int receive_responses(neo4j_session_t *session, const unsigned int *condition)
         }
     }
 
-    return failure? 1 : 0;
+    if (interruptable && interrupted(session))
+    {
+        return 1;
+    }
+
+    assert(!failure || session->inflight_requests == 0);
+    return failure? 2 : 0;
 }
 
 
+/**
+ * Send IGNORED to all queued requests.
+ *
+ * @internal
+ *
+ * This will also generate IGNORED for all inflight requests, so this
+ * method should only be called when there are no inflight requests or
+ * when a terminal error has occured and the connection will be closed.
+ *
+ * @param [session] The session.
+ * @return 0 on success, -1 if an error occurs (errno will be set).
+ */
 int drain_queued_requests(neo4j_session_t *session)
 {
     assert(session != NULL);
@@ -423,10 +659,23 @@ int drain_queued_requests(neo4j_session_t *session)
 }
 
 
+/**
+ * Add a queued request.
+ *
+ * @internal
+ *
+ * The returned pointer will be to a request struct already added to the tail
+ * of the queue. It MUST be populated with valid attributes before any other
+ * session methods are invoked.
+ *
+ * @param [session] The session.
+ * @return The queued request, which MUST be populated with valid attributes,
+ *         or `NULL` if an error occurs (errno will be set).
+ */
 struct neo4j_request *new_request(neo4j_session_t *session)
 {
     assert(session != NULL);
-    const neo4j_config_t *config = session->config;
+    const neo4j_config_t *config = neo4j_session_config(session);
 
     if (session->failed)
     {
@@ -457,6 +706,13 @@ struct neo4j_request *new_request(neo4j_session_t *session)
 }
 
 
+/**
+ * Pop a request off the head of the queue.
+ *
+ * @internal
+ *
+ * @param [session] The session.
+ */
 void pop_request(neo4j_session_t* session)
 {
     assert(session != NULL);
@@ -477,35 +733,114 @@ void pop_request(neo4j_session_t* session)
 }
 
 
-int initialize(neo4j_session_t *session)
+struct init_cdata
+{
+    neo4j_session_t *session;
+    int error;
+};
+
+
+int initialize(neo4j_session_t *session, unsigned int attempts)
 {
     assert(session != NULL);
-    const neo4j_config_t *config = session->config;
+    neo4j_config_t *config = neo4j_session_config(session);
     const char *client_id = config->client_id;
-    const char *principal = config->username? config->username : "";
-    const char *credentials = config->password? config->password : "";
 
-    struct neo4j_request *req = new_request(session);
-    if (req == NULL)
+    struct init_cdata cdata = { .session = session, .error = 0 };
+
+    const char *username = (config->username != NULL)? config->username : "";
+    const char *password = (config->password != NULL)? config->password : "";
+
+    if (attempts > 0 || config->auth_reattempt_callback == NULL ||
+            config->password != NULL || config->attempt_empty_password)
+    {
+        struct neo4j_request *req = new_request(session);
+        if (req == NULL)
+        {
+            return -1;
+        }
+
+        req->type = NEO4J_INIT_MESSAGE;
+        req->_argv[0] = neo4j_string(client_id);
+        neo4j_map_entry_t auth_token[3] =
+            { neo4j_map_entry("scheme", neo4j_string("basic")),
+              neo4j_map_entry("principal", neo4j_string(username)),
+              neo4j_map_entry("credentials", neo4j_string(password)) };
+        req->_argv[1] = neo4j_map(auth_token, 3);
+        req->argv = req->_argv;
+        req->argc = 2;
+        req->receive = initialize_callback;
+        req->cdata = &cdata;
+
+        neo4j_log_trace(session->logger, "enqu INIT{\"%s\"} (%p) in %p",
+                client_id, (void *)req, (void *)session);
+
+        if (neo4j_session_sync(session, NULL))
+        {
+            return -1;
+        }
+
+        if (cdata.error == 0)
+        {
+            return 0;
+        }
+
+        assert(cdata.error == NEO4J_INVALID_CREDENTIALS ||
+                cdata.error == NEO4J_AUTH_RATE_LIMIT);
+
+        if (config->auth_reattempt_callback == NULL)
+        {
+            errno = cdata.error;
+            return -1;
+        }
+
+        ++attempts;
+    }
+
+    char host[NEO4J_MAXHOSTLEN];
+    if (describe_host(host, sizeof(host),
+            session->connection->hostname, session->connection->port))
     {
         return -1;
     }
-    req->type = NEO4J_INIT_MESSAGE;
-    req->_argv[0] = neo4j_string(client_id);
-    neo4j_map_entry_t auth_token[3] =
-        { neo4j_map_entry("scheme", neo4j_string("basic")),
-          neo4j_map_entry("principal", neo4j_string(principal)),
-          neo4j_map_entry("credentials", neo4j_string(credentials)) };
-    req->_argv[1] = neo4j_map(auth_token, 3);
-    req->argv = req->_argv;
-    req->argc = 2;
-    req->receive = initialize_callback;
-    req->cdata = session;
 
-    neo4j_log_trace(session->logger, "enqu INIT{\"%s\"} (%p) in %p",
-            client_id, (void *)req, (void *)session);
+    char username_buf[NEO4J_MAXUSERNAMELEN];
+    strncpy(username_buf, username, sizeof(username_buf));
+    char password_buf[NEO4J_MAXPASSWORDLEN];
+    strncpy(password_buf, password, sizeof(password_buf));
 
-    return neo4j_session_sync(session, NULL);
+    int r = config->auth_reattempt_callback(
+            config->auth_reattempt_callback_userdata, host, attempts,
+            cdata.error, username_buf, sizeof(username_buf),
+            password_buf, sizeof(password_buf));
+    if (r < 0)
+    {
+        return -1;
+    }
+    else if (r > 0)
+    {
+        errno = cdata.error;
+        if (cdata.error == 0)
+        {
+            neo4j_log_error(session->logger,
+                    "authentication callback returned NEO4J_AUTHENTICATION_FAIL"
+                    " before first authentication attempt (in %p)",
+                    (void *)session);
+            errno = NEO4J_UNEXPECTED_ERROR;
+        }
+        return -1;
+    }
+
+    if (neo4j_config_set_username(config, username_buf))
+    {
+        return -1;
+    }
+    if (neo4j_config_set_password(config, password_buf))
+    {
+        return -1;
+    }
+
+    return initialize(session, attempts);
 }
 
 
@@ -513,10 +848,28 @@ int initialize_callback(void *cdata, neo4j_message_type_t type,
         const neo4j_value_t *argv, uint16_t argc)
 {
     assert(cdata != NULL);
-    neo4j_session_t *session = (neo4j_session_t *)cdata;
+    neo4j_session_t *session = ((struct init_cdata *)cdata)->session;
+
+    char description[128];
 
     if (type == NEO4J_SUCCESS_MESSAGE)
     {
+        snprintf(description, sizeof(description),
+                "SUCCESS in %p (response to INIT)", (void *)session);
+        const neo4j_value_t *metadata = neo4j_validate_metadata(argv, argc,
+                description, session->logger);
+        if (metadata == NULL)
+        {
+            return -1;
+        }
+        if (neo4j_log_is_enabled(session->logger, NEO4J_LOG_TRACE))
+        {
+            neo4j_metadata_log(session->logger, NEO4J_LOG_TRACE, description,
+                    *metadata);
+        }
+        neo4j_value_t ce = neo4j_map_get(*metadata, "credentials_expired");
+        session->credentials_expired =
+                (neo4j_type(ce) == NEO4J_BOOL && neo4j_bool_value(ce));
         return 0;
     }
 
@@ -531,10 +884,8 @@ int initialize_callback(void *cdata, neo4j_message_type_t type,
     }
 
     // handle failure
-    char description[128];
     snprintf(description, sizeof(description),
-            "FAILURE message received in %p (in response to INIT)",
-            (void *)session);
+            "FAILURE in %p (response to INIT)", (void *)session);
     const neo4j_value_t *metadata = neo4j_validate_metadata(argv, argc,
             description, session->logger);
     if (metadata == NULL)
@@ -542,27 +893,51 @@ int initialize_callback(void *cdata, neo4j_message_type_t type,
         return -1;
     }
 
-    const neo4j_config_t *config = session->config;
-    const char *code;
-    const char *message;
+    if (neo4j_log_is_enabled(session->logger, NEO4J_LOG_TRACE))
+    {
+        neo4j_metadata_log(session->logger, NEO4J_LOG_TRACE, description,
+                *metadata);
+    }
+
+    const neo4j_config_t *config = neo4j_session_config(session);
+    struct neo4j_failure_details details;
     neo4j_mpool_t mpool =
         neo4j_mpool(config->allocator, config->mpool_block_size);
-    if (neo4j_meta_failure_details(&code, &message, *metadata, &mpool,
+    if (neo4j_meta_failure_details(&details, *metadata, &mpool,
                 description, session->logger))
     {
         return -1;
     }
 
-    if (strcmp(code, "Neo.ClientError.Security.EncryptionRequired") == 0)
+    int result = -1;
+
+    if (strcmp("Neo.ClientError.Security.EncryptionRequired",
+            details.code) == 0)
     {
         errno = NEO4J_SERVER_REQUIRES_SECURE_CONNECTION;
-        return -1;
+        goto cleanup;
+    }
+    if (strcmp("Neo.ClientError.Security.Unauthorized", details.code) == 0)
+    {
+        ((struct init_cdata *)cdata)->error = NEO4J_INVALID_CREDENTIALS;
+        result = 0;
+        goto cleanup;
+    }
+    if (strcmp("Neo.ClientError.Security.AuthenticationRateLimit",
+            details.code) == 0)
+    {
+        ((struct init_cdata *)cdata)->error = NEO4J_AUTH_RATE_LIMIT;
+        result = 0;
+        goto cleanup;
     }
 
-    neo4j_log_error(session->logger, "session initalization failed: %s",
-            message);
+    neo4j_log_error(session->logger, "session initialization failed: %s",
+            details.message);
     errno = NEO4J_UNEXPECTED_ERROR;
-    return -1;
+
+cleanup:
+    neo4j_mpool_drain(&mpool);
+    return result;
 }
 
 
@@ -614,54 +989,6 @@ int ack_failure_callback(void *cdata, neo4j_message_type_t type,
 }
 
 
-int reset(neo4j_session_t *session)
-{
-    assert(session != NULL);
-
-    struct neo4j_request *req = new_request(session);
-    if (req == NULL)
-    {
-        return -1;
-    }
-    req->type = NEO4J_RESET_MESSAGE;
-    req->argc = 0;
-    req->receive = reset_callback;
-    req->cdata = session;
-
-    neo4j_log_trace(session->logger, "enqu RESET (%p) in %p",
-            (void *)req, (void *)session);
-
-    return neo4j_session_sync(session, NULL);
-}
-
-
-int reset_callback(void *cdata, neo4j_message_type_t type,
-        const neo4j_value_t *argv, uint16_t argc)
-{
-    assert(cdata != NULL);
-    neo4j_session_t *session = (neo4j_session_t *)cdata;
-
-    if (type == NEO4J_IGNORED_MESSAGE)
-    {
-        // only when draining after connection close
-        return 0;
-    }
-    if (type != NEO4J_SUCCESS_MESSAGE)
-    {
-        neo4j_log_error(session->logger,
-                "unexpected %s message received in %p"
-                " (expected SUCCESS in response to RESET)",
-                neo4j_message_type_str(type), (void *)session);
-        errno = EPROTO;
-        return -1;
-    }
-
-    neo4j_log_trace(session->logger, "RESET complete in %p",
-            (void *)session);
-    return 0;
-}
-
-
 int neo4j_session_run(neo4j_session_t *session, neo4j_mpool_t *mpool,
         const char *statement, neo4j_value_t params,
         neo4j_response_recv_t callback, void *cdata)
@@ -672,10 +999,17 @@ int neo4j_session_run(neo4j_session_t *session, neo4j_mpool_t *mpool,
     REQUIRE(neo4j_type(params) == NEO4J_MAP || neo4j_is_null(params), -1);
     REQUIRE(callback != NULL, -1);
 
+    if (neo4j_atomic_bool_set(&(session->processing), true))
+    {
+        errno = NEO4J_SESSION_BUSY;
+        return -1;
+    }
+
+    int err = -1;
     struct neo4j_request *req = new_request(session);
     if (req == NULL)
     {
-        return -1;
+        goto cleanup;
     }
     req->type = NEO4J_RUN_MESSAGE;
     req->_argv[0] = neo4j_string(statement);
@@ -689,7 +1023,11 @@ int neo4j_session_run(neo4j_session_t *session, neo4j_mpool_t *mpool,
     neo4j_log_trace(session->logger, "enqu RUN{\"%s\"} (%p) in %p",
             statement, (void *)req, (void *)session);
 
-    return 0;
+    err = 0;
+
+cleanup:
+    neo4j_atomic_bool_set(&(session->processing), false);
+    return err;
 }
 
 
@@ -700,10 +1038,17 @@ int neo4j_session_pull_all(neo4j_session_t *session, neo4j_mpool_t *mpool,
     REQUIRE(mpool != NULL, -1);
     REQUIRE(callback != NULL, -1);
 
+    if (neo4j_atomic_bool_set(&(session->processing), true))
+    {
+        errno = NEO4J_SESSION_BUSY;
+        return -1;
+    }
+
+    int err = -1;
     struct neo4j_request *req = new_request(session);
     if (req == NULL)
     {
-        return -1;
+        goto cleanup;
     }
 
     req->type = NEO4J_PULL_ALL_MESSAGE;
@@ -716,7 +1061,11 @@ int neo4j_session_pull_all(neo4j_session_t *session, neo4j_mpool_t *mpool,
     neo4j_log_trace(session->logger, "enqu PULL_ALL (%p) in %p",
             (void *)req, (void *)session);
 
-    return 0;
+    err = 0;
+
+cleanup:
+    neo4j_atomic_bool_set(&(session->processing), false);
+    return err;
 }
 
 
@@ -727,10 +1076,17 @@ int neo4j_session_discard_all(neo4j_session_t *session, neo4j_mpool_t *mpool,
     REQUIRE(mpool != NULL, -1);
     REQUIRE(callback != NULL, -1);
 
+    if (neo4j_atomic_bool_set(&(session->processing), true))
+    {
+        errno = NEO4J_SESSION_BUSY;
+        return -1;
+    }
+
+    int err = -1;
     struct neo4j_request *req = new_request(session);
     if (req == NULL)
     {
-        return -1;
+        goto cleanup;
     }
 
     req->type = NEO4J_DISCARD_ALL_MESSAGE;
@@ -743,5 +1099,9 @@ int neo4j_session_discard_all(neo4j_session_t *session, neo4j_mpool_t *mpool,
     neo4j_log_trace(session->logger, "enqu DISCARD_ALL (%p) in %p",
             (void *)req, (void *)session);
 
-    return 0;
+    err = 0;
+
+cleanup:
+    neo4j_atomic_bool_set(&(session->processing), false);
+    return err;
 }
